/*************************************************************************
 * source code from ogmnav.c * Author: George Shuklin, gs@shounen.ru (http://shounen.ru/docs/ogm/)
 * No licence, PUBLIC DOMAIN
 *************************************************************************/

 /** Russian, CP-1251 codepage *******************************************
 *     ogmnav.c * :  , gs@shounen.ru (http://shounen.ru/docs/ogm/)
 *  ,   (public domain)
 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#define _test_ printf

#define PART_HEADER_SIZE 0x1B

#define CONTINUED_FLAG  0x1
#define BoS_FLAG    0x2
#define EoS_FLAG    0x4

#define CharsToInt32(table,offset)  (table[offset]+(table[offset+1]<<8)+(table[offset+2]<<16)+(table[offset+3]<<24))
#define CharsToInt64H(table,offset) (CharsToInt32(table,offset+4))
#define CharsToInt64L(table,offset) (CharsToInt32(table,offset))

#if UINT_MAX >= 0xffffffff
typedef unsigned int crc32_t;
#else
typedef unsigned long crc32_t;
#endif

#define ViewData 0 //    

extern unsigned char ogg_corr[];
//unsigned char ogg_corr_N[32]; //HEX DUMP ogg-comment TRACKNUMBER=123456789012
unsigned char ogg_corr_N[]={0x01,0x00,0x00,0x00,0x18,0x00,0x00,0x00,0x54,0x52,0x41,0x43,0x4b,0x4e,0x55,0x4d,0x42,0x45,0x52,0x3d,0x31,0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,0x30,0x31,0x32};


/*
      OGM (OGG)
*/
int oggnav( char* fname ,int mod ){
// fname - filename .ogg
// mod - 0 - check/view
// mod - 1 - correct ogg page crc
// mod - 2 - add ogg-comment


  FILE* file=NULL;
  FILE* file_new=NULL;

  crc32_t crc32;
  crc32_t crc32_file;
  unsigned char part_header[PART_HEADER_SIZE]; /*    (  segment_table)*/
  unsigned char *segment_table=NULL; /*   segment_table,    page_segments*/
  const unsigned char capture_pattern[4]={0x4f, 0x67, 0x67, 0x53}; /*  , 'OggS'*/
  int page_start=0;      /*    "" () */
  int page_start_prev=0; /*    "" */
  int  page_data_size;    /*     */
  int c; /*    segment_table*/
  int stream_counter=0; /*      */
  int page_counter=0; /*     */
  unsigned char *databuf=0; /*   */
  char fname_new[264];

//_test_("MODE :[%d]\n\n",mod);

  file = fopen( fname, "rb+" );
  if( !file ){
    printf( "!!! Error opening file: %s\n", fname );
    return -1;
  }

  if (mod==2)
  { memset(&fname_new, 0, sizeof(fname_new));
    strcpy(fname_new,fname);
    strcat(fname_new,"~.tmp");
    file_new = fopen( fname_new, "wb" );
    if( !file_new ){
      printf( "!!! Error opening new_file: %s\n", fname_new );
      fclose( file );
      return -1;
    }

  }

  while( fread( part_header, 1, sizeof(part_header), file ) )
  {
    crc32 = 0x00000000;
    if( memcmp( part_header, capture_pattern, 4 ) != 0 ){/*    ,   ,  */
      printf( "!!! --- sync lost at offset %d (0x%X), must be 'OggS', is '%c%c%c%c' ---\n",
        page_start, page_start, part_header[0], part_header[1], part_header[2], part_header[3] );
      if( segment_table ) free( segment_table );
      fclose( file );
      if(file_new != NULL) fclose( file_new );
      return -1;
    }
    if( part_header[4] != 0 ){
      printf( "!!! --- unknown version of page header structure, must be 0, is %d ---\n", part_header[5] );
      if( segment_table ) free( segment_table );
      fclose( file );
      if(file_new != NULL) fclose( file_new );
      return -1;
    }

    /*ok,   ,  */
    page_counter++;
    memcpy(&crc32_file,part_header+22,4);
    
//    if ((CharsToInt32(part_header,18))<2) ViewData=1; else  ViewData=0;

#if ViewData
    printf( "\n....... page #%d of stream #%d, at %d (0x%X)\n", CharsToInt32(part_header,18), CharsToInt32(part_header,14), page_start, page_start );

    if( part_header[5] != 0 ){
      printf( "... This page: " );
      if( part_header[5] & CONTINUED_FLAG ) printf( "contain a continued packet " );
      if( part_header[5] & BoS_FLAG       ) printf( "is the Begin of Stream #%d", stream_counter++ );
      if( part_header[5] & EoS_FLAG       ) printf( "is the End of Stream" );
      printf( "\n" );
    }

    printf("... header_type_flag          = 0x%X\n", (int)part_header[5] );
    printf("... absolute_granule_position = 0x%X%X\n", CharsToInt64H( part_header, 6 ), CharsToInt64L( part_header, 6 ) );
    printf("... page_checksum (Char2Int)  = 0x%X\n", CharsToInt32( part_header, 22 ) );
    //printf("... page_checksum (Int32)     = 0x%X\n", crc32_file);
    printf("... page_segments             = %d\n", (int)part_header[26] );
#endif

    part_header[22]=0; part_header[23]=0; part_header[24]=0; part_header[25]=0; // crc32 
    if (mod==1) crc32 = get_crc(crc32, part_header, sizeof(part_header));
        /*  segment_table*/
    if( part_header[26] == 0 )
    {
#if ViewData
 printf( "....... this page do not contain segment_table and data\n" );
#endif

          continue;
    }
    segment_table =malloc( (int)part_header[26] );
    if( !segment_table )
    {
      printf( "!!! !!! Error allocating %d bytes of memory. Program terminating\n", part_header[26] );
      fclose( file );
      if(file_new != NULL) fclose( file_new );
      return -1;
    }
    if( !fread( segment_table, 1, (int)part_header[26], file ) )
    {
      if( segment_table ) free( segment_table );
      fclose( file );
      if(file_new != NULL) fclose( file_new );
      printf( "!!! --- Error reading segment_table (%d bytes), offset:%d (0x%d) ---\n", part_header[26], (int)ftell(file), (int)ftell(file) );
      return -1;
    }
    if (mod==1) crc32 = get_crc(crc32, segment_table, (int)part_header[26]);

    /* segment_table*/
    page_data_size = 0;
    for( c=0; c<part_header[26]; c++ )
    {
      page_data_size += segment_table[c];
    }
#if ViewData
        printf( "... page_data_size            = %d\n",page_data_size);
        printf("... common_segment_size       = %d\n", page_data_size+part_header[26]+PART_HEADER_SIZE );
#endif

    databuf = malloc( page_data_size );
    if( !databuf ){
      printf("!!! !!! Error allocating %d bytes of memory. Program terminating\n", page_data_size);
      if(segment_table) free(segment_table);
      fclose(file);
      if(file_new != NULL) fclose( file_new );
      return -1;
    }
    if(!fread(databuf,1,page_data_size,file)){
      printf("!!! --- Error reading page content\n");
      if(segment_table) free(segment_table);
      fclose(file);
      if(file_new != NULL) fclose( file_new );
      return -1;
    }
    if (mod==1) crc32 = get_crc(crc32, databuf, page_data_size);

    if (mod==2)
    {
    	int comment_flag=0; //   ,   

			int offset=0;/*       */
			int base=0, comment_base=0;

					if(!(!databuf || (page_data_size<8) || (part_header[5]&CONTINUED_FLAG))) /*     */
					{
			
			    	if (databuf[0]==0x03) //    
						{
							comment_flag=1;
			        comment_base=11+CharsToInt32(databuf,7); //"vorbis"+4+    ,   "Created using"
			        offset=comment_base+4;
			          
			        for(c=0;c<CharsToInt32(databuf,comment_base);c++)
			        {
			            offset+=4+CharsToInt32(databuf,offset);
			        }
						}
					}

			if (comment_flag)
			{

  			unsigned char *newdatabuf=0; /*   */
    		int newst=comment_base+32-offset;

    		newdatabuf = malloc( page_data_size+part_header[26]+PART_HEADER_SIZE+32);

    		base=0;
        // 
    		memcpy(newdatabuf,part_header,sizeof(part_header));
    		base+=sizeof(part_header);

        if ((comment_base-offset)!=32) //    !=32 (28 + 4 )
        {
          for( c=0; c<part_header[26]; c++ )// segment_table
          {
            if ((((int)segment_table[c]+newst)>0) && (((int)segment_table[c]+newst)<256))
            {
            	segment_table[c]+=newst;
            	break;
            }
            else
            {
            	if (c<(part_header[26]-1))
            	{
            	  segment_table[c]=0xFF;
            	  newst-=segment_table[c+1];
            	}
            	else
            		if (newst<256)
            	  {
            	  	segment_table[c]=newst;
            	  }
            	  else
            	  {  printf("!!! --- Error set segment_table \n");
                   fclose(file);
                   if(file_new != NULL) fclose( file_new );
                   return -1;
            	  }
            }
          }
        }

//comment_base -    ( - )
//offset -   

        //  
    		memcpy(newdatabuf+base,segment_table,(int)part_header[26]);
    		base+=(int)part_header[26];


        //    ( )
    		memcpy(newdatabuf+base,databuf,comment_base);
    		base+=comment_base;

        // 
    		memcpy(ogg_corr_N+8,ogg_corr,24);
        //  
    		memcpy(newdatabuf+base,ogg_corr_N,32);
    		base+=32;

        //  
    		memcpy(newdatabuf+base,databuf+offset,page_data_size-offset);
    		base+=page_data_size-offset;

        crc32 = get_crc(crc32, newdatabuf, base); // crc32
        memcpy(newdatabuf+22,&crc32,4);

        c=fwrite(newdatabuf, 1, base, file_new);
        if (c!=base)
        {
          printf( " Write - FAILED L:%d!\n",__LINE__);
          fclose( file );
          fclose( file_new );
          free(newdatabuf);
          return -1;
        }
        free(newdatabuf);
      }//if comments block
      else
      {//if no comments

        //#!#!#!#!#!#!#!#!# fwrite
        
        memcpy(part_header+22,&crc32_file,4);
        
        c=fwrite(part_header, 1, sizeof(part_header), file_new);
        if (c!=sizeof(part_header))
        {
          printf( " Write - FAILED L:%d!\n",__LINE__);
          fclose( file );
          fclose( file_new );
          return -1;
        }

        //#!#!#!#!#!#!#!#!# fwrite
        c=fwrite(segment_table, 1, (int)part_header[26], file_new);
        if (c!=(int)part_header[26])
        {
          printf( " Write - FAILED L:%d!\n",__LINE__);
          fclose( file );
          fclose( file_new );
          return -1;
        }

        //#!#!#!#!#!#!#!#!# fwrite
        c=fwrite(databuf, 1, page_data_size, file_new);
        if (c!=page_data_size)
        {
          printf( " Write - FAILED L:%d!\n",__LINE__);
          fclose( file );
          fclose( file_new );
          return -1;
        }
      }//if no comments
    }
    free( segment_table );
    segment_table = NULL;

    free(databuf);
    databuf = NULL;

    page_start_prev=page_start;
    page_start = ftell( file );

    if (mod==1)
    {
      if( crc32 != crc32_file )
      {/*  CRC32 */
#if ViewData
        printf("... Calculated CRC            = 0x%X - !!! FAILED.\n", crc32);
#endif
        c=fseek (file , page_start_prev+22, SEEK_SET);

        c=fwrite(&crc32, 1, 4, file);

        if (c==4)
        {
#if ViewData
           printf( " Corrected - OK.\n");
#endif
        }
        else
        {
          printf( " Corrected - FAILED !\n");
        }

        c=fseek (file , page_start, SEEK_SET);
      }
      else
      {
#if ViewData
        printf("... Calculated CRC            = 0x%X - checked.\n", crc32);
#endif
      }
    }
  }

  if(file != NULL)
  {
    fclose( file );
    file = NULL;
  }

  if(file_new != NULL)
  {
    fclose( file_new );
    file_new = NULL;
  }

#if ViewData

  printf("\n    --- EoF ---\n");
  printf(" \n\nStatistic: %d streams, %d pages\n", stream_counter, page_counter );
#endif

  if (mod==2)
  {

   if(remove(fname) != 0) 
   {
      printf("!!! --- Error: unable to delete origin file.\n");
   }

   if(rename(fname_new, fname) != 0) 
   {
      printf("!!! --- Error: unable to rename temp file.\n");
   }
  }


  return 0;
}
